import json
from django.contrib.auth.decorators import login_required
from django.views.generic import View
from django.utils.translation import gettext_lazy as _
from django.apps import apps
from django.http import HttpResponse
from accrete.views import TenantRequiredMixin
from accrete.models import AccessGroup
from accrete.contrib.ui.filter import Filter
from accrete.contrib.ui.response import WindowResponse, ModalResponse


class TenantView(TenantRequiredMixin, View):

    """
    Base View that handles displaying access denied messages
    to the user if member/tenant groups are missing.
    """

    def handle_tenant_group_not_set(self):
        if not self._is_htmx():
            return self._access_denied_page_response()
        return self._access_denied_modal_response()

    def handle_member_group_not_set(self):
        if not self._is_htmx():
            return self._access_denied_page_response()
        return self._access_denied_modal_response()

    def _access_denied_page_response(self):
        return WindowResponse(
            title=str(_('Access Denied')),
            overview_template='mirox/base/group_not_set.html',
            context=dict(groups=self._get_group_data()),
            is_centered=True
        ).response(self.request)

    def _access_denied_modal_response(self):
        res = ModalResponse(
            template='mirox/base/group_not_set_modal.html',
            title=str(_('Access Denied')),
            modal_id='group-missing-modal',
            context=dict(groups=self._get_group_data())
        ).response(self.request)
        res.headers['HX-Reswap'] = 'none'
        res.headers['HX-Push-Url'] = 'false'
        return res

    def _get_group_data(self) -> dict:
        data = {}
        tenant_groups, member_groups = self._flat_groups()
        if tenant_groups:
            data.update(tenant_groups=[])
            access_groups = AccessGroup.objects.filter(
                code__in=tenant_groups,
                apply_on='tenant'
            ).all()
            group_data = {item[0]: item[1] for item in access_groups.values_list('code', 'name')}
            for group in self.TENANT_GROUPS:
                if isinstance(group, tuple):
                    data['tenant_groups'].append(' & '.join([group_data.get(g, g) for g in group]))
                else:
                    data['tenant_groups'].append(group_data.get(group, group))
        if member_groups:
            data.update(member_groups=[])
            access_groups = AccessGroup.objects.filter(
                code__in=member_groups,
                apply_on='member'
            ).all()
            group_data = {item[0]: item[1] for item in access_groups.values_list('code', 'name')}
            for group in self.MEMBER_GROUPS:
                if isinstance(group, tuple):
                    data['member_groups'].append(' & '.join([group_data.get(g, g) for g in group]))
                else:
                    data['member_groups'].append(group_data.get(group, group))
        return data

    def _flat_groups(self) -> tuple[list[str], list[str]]:
        def group_list(g):
            if isinstance(g, str):
                return [g]
            elif isinstance(g, tuple):
                return [x for x in g]
            return []
        tenant_groups = []
        member_groups = []
        for group in self.TENANT_GROUPS:
            tenant_groups.extend(group_list(group))
        for group in self.MEMBER_GROUPS:
            member_groups.extend(group_list(group))
        return tenant_groups, member_groups

    def _is_htmx(self):
        return self.request.headers.get('HX-Request', 'false') == 'true'


@login_required
def params(request, model: str):
    app_label, model_name = model.split('.')
    Model = apps.get_model(app_label, model_name)
    return HttpResponse(Filter(Model, request.GET).query_params())


@login_required
def set_filter_input(request, model: str):
    app_label, model_name = model.split('.')
    Model = apps.get_model(app_label, model_name)
    lookup = request.GET.get('lookup')
    return HttpResponse(Filter(Model, request.GET).query_input(lookup))


@login_required
def filter_add_query(request, model: str):
    app_label, model_name = model.split('.')
    Model = apps.get_model(app_label, model_name)
    query = json.loads(request.GET.get('q', '[]'))
    lookup = request.GET.get('filter_lookup')
    value = request.GET.get('filter_input')
    query.append({lookup: value})
    query = json.dumps(query)
    query_dict = request.GET.copy()
    query_dict.update(q=query)
    return HttpResponse(Filter(Model, query_dict).query_tags())



import logging

from django.db import models

from accrete.tenant import get_member, get_tenant
from accrete.utils.log import log_time
from .models import Log, LogConfig
from . import helper, queries, config

_logger = logging.getLogger(__name__)


@log_time
def create_log(sender, **kwargs):
    if (
        sender._meta.object_name == 'Migration'
        and not config.ACCRETE_LOG_RUN_IN_MIGRATION
    ):
        return

    def _create_log():
        member = get_member()
        user = member and member.user or None
        object_id = instance.id
        log_rows = queries.current_log_state(model, object_id)
        log_state, log_info = helper.log_state_to_dict(log_rows)
        instance_state = helper.get_instance_state(instance)
        diff = dict(log_state.items() ^ instance_state.items())
        logs_to_create = []
        for field, value in instance_state.items():
            if field not in diff.keys():
                continue
            value_type = helper.internal_type_to_log_type(instance._meta.get_field(field))
            if value_type is None:
                continue
            if value is None:
                new_value = None
            else:
                new_value = str(value)
            log = Log(
                model=model,
                field=field,
                object_id=instance.id,
                old_value_type=log_info.get(field, {}).get('new_value_type', ''),
                old_value=log_state.get(field),
                new_value_type=helper.internal_type_to_log_type(instance._meta.get_field(field)),
                new_value=new_value,
                user=user,
                tenant=get_tenant()
            )
            logs_to_create.append(log)
        Log.objects.bulk_create(logs_to_create)

    instance: models.Model = kwargs.get('instance')
    model = f'{instance._meta.app_label}.{instance._meta.model_name}'
    log_config = LogConfig.objects.filter(model=model).first()
    if not log_config:
        return
    try:
        _create_log()
    except Exception as e:
        _logger.exception(e)
        if log_config.ignore_errors:
            pass
        else:
            raise e



# Generated by Django 5.1.1 on 2025-01-10 07:56

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('log', '0001_initial'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.AddField(
            model_name='log',
            name='user',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='logs', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='log',
            name='new_value_type',
            field=models.CharField(choices=[('fk', 'Foreign Key'), ('int', 'Integer'), ('float', 'Float'), ('decimal', 'Decimal'), ('bool', 'Boolean'), ('str', 'String'), ('date', 'Date'), ('datetime', 'Date Time')], max_length=100, verbose_name='New Value Type'),
        ),
        migrations.AlterField(
            model_name='log',
            name='old_value_type',
            field=models.CharField(choices=[('fk', 'Foreign Key'), ('int', 'Integer'), ('float', 'Float'), ('decimal', 'Decimal'), ('bool', 'Boolean'), ('str', 'String'), ('date', 'Date'), ('datetime', 'Date Time')], max_length=100, verbose_name='Old Value Type'),
        ),
    ]



# Generated by Django 5.1.7 on 2025-03-16 19:37

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('auth', '0012_alter_user_first_name_max_length'),
        ('user', '0004_user_login_alter_user_email_user_email_or_login_set'),
    ]

    operations = [
        migrations.RemoveConstraint(
            model_name='user',
            name='email_or_login_set',
        ),
        migrations.AddConstraint(
            model_name='user',
            constraint=models.CheckConstraint(condition=models.Q(('email__isnull', False), ('login__isnull', False), _connector='OR'), name='email_or_login_set', violation_error_message='email or login must be set'),
        ),
    ]



from django.contrib import admin
from .models import Theme


class ThemeAdmin(admin.ModelAdmin):
    model = Theme
    list_display = ('pk', 'user', 'tenant')
    search_fields = ('pk', 'user', 'tenant')


admin.site.register(Theme, ThemeAdmin)



import logging
from functools import wraps
from timeit import default_timer as timer

_logger = logging.getLogger(__name__)


def log_time(f):
    @wraps(f)
    def decorator(*args, **kwargs):
        start = timer()
        result = f(*args, **kwargs)
        stop = timer()
        _logger.info(f'{f.__module__}.{f.__name__} Duration: {(stop - start)} Seconds')
        return result
    return decorator



# Generated by Django 5.1.7 on 2025-03-19 17:53

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('auth', '0012_alter_user_first_name_max_length'),
        ('user', '0006_remove_user_email_or_login_set_user_is_managed_and_more'),
    ]

    operations = [
        migrations.AddConstraint(
            model_name='user',
            constraint=models.CheckConstraint(condition=models.Q(models.Q(('is_managed', True), ('login__isnull', False)), ('is_managed', False), _connector='OR'), name='managed_login', violation_error_message='Managed users must have a login'),
        ),
    ]



import colorsys

from django.conf import settings
from django.db import models
from django.utils.translation import gettext_lazy as _
from django.template.loader import render_to_string


class Theme(models.Model):

    class Meta:
        verbose_name = _('Theme')
        verbose_name_plural = _('Themes')
        db_table = 'accrete_theme'
        ordering = ['pk']
        constraints = [
            models.CheckConstraint(
                name='check_user_or_tenant',
                violation_error_message=_(
                    'A theme must be assigned either to a user or a tenant.'
                ),
                condition=(
                    models.Q(user__isnull=True) ^ models.Q(tenant__isnull=True)
                )
            )
        ]

    user = models.ForeignKey(
        verbose_name=_('User'),
        to=settings.AUTH_USER_MODEL,
        on_delete=models.CASCADE,
        related_name='user_themes',
        null=True,
        blank=True
    )

    tenant = models.ForeignKey(
        verbose_name=_('Tenant'),
        to='accrete.Tenant',
        on_delete=models.CASCADE,
        related_name='tenant_themes',
        null=True,
        blank=True
    )

    force_tenant_theme = models.BooleanField(
        verbose_name=_('Force Tenant Theme'),
        default=False,
        help_text=_('Apply this theme regardless of user preferences')
    )

    color_primary = models.CharField(
        verbose_name=_('Primary Color'),
        max_length=7,
        default='#46C68D'
    )

    color_success = models.CharField(
        verbose_name=_('Success Color'),
        max_length=7,
        default='#46C68D'
    )

    color_link = models.CharField(
        verbose_name=_('Link Color'),
        max_length=7,
        default='#65D0FE'
    )

    color_warning = models.CharField(
        verbose_name=_('Warning Color'),
        max_length=7,
        default='#FEB600'
    )

    color_danger = models.CharField(
        verbose_name=_('Danger Color'),
        max_length=7,
        default='#FF6584'
    )

    theme_markup = models.TextField(
        verbose_name=_('Theme Markup'),
        null=True,
        blank=True
    )

    def generate_markup(self) -> str:
        def hex_to_hsl(hex_color: str) -> dict:
            hex_color = hex_color.removeprefix('#')
            rgb = tuple(int(hex_color[i:i+2], 16) for i in (0, 2, 4))
            rgb = [x/255.0 for x in rgb]
            h, l, s = colorsys.rgb_to_hls(*rgb)

            return {
                'h': round(h * 360),
                's': round(s * 100),
                'l': round(l * 100)
            }
        ctx = {
            'primary': hex_to_hsl(self.color_primary),
            'success': hex_to_hsl(self.color_success),
            'link': hex_to_hsl(self.color_link),
            'warning': hex_to_hsl(self.color_warning),
            'danger': hex_to_hsl(self.color_danger)
        }
        return render_to_string('ui/custom_theme.html', ctx)

    def save(
        self,
        force_insert=False,
        force_update=False,
        using=None,
        update_fields=None,
    ):
        self.theme_markup = self.generate_markup()
        super().save(force_insert, force_update, using, update_fields)



# Generated by Django 5.1.1 on 2025-01-12 09:24

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('log', '0004_logconfig_logconfigfield'),
    ]

    operations = [
        migrations.AddField(
            model_name='logconfig',
            name='exclude_fields',
            field=models.BooleanField(default=False, help_text='If set, Log Configuration Fields will be excluded from logging. Otherwise only configured fields will be logged.', verbose_name='Exclude Fields'),
        ),
        migrations.AlterField(
            model_name='logconfig',
            name='ignore_errors',
            field=models.BooleanField(default=False, help_text='If set, exceptions during log creation will be ignored', verbose_name='Ignore Errors'),
        ),
        migrations.AlterField(
            model_name='logconfigfield',
            name='log_config',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='fields', to='log.logconfig', verbose_name='Log Configuration'),
        ),
    ]



from django.apps import AppConfig
from django.db.models.signals import post_save


class LogConfig(AppConfig):
    default_auto_field = 'django.db.models.BigAutoField'
    name = 'accrete.contrib.log'

    def ready(self):
        from . import signals
        post_save.connect(
            signals.create_log, weak=False, dispatch_uid="accrete.contrib.log"
        )



import django.apps
from django.contrib import admin
from django import forms
from . import models

installed_models = django.apps.apps.get_models(
    include_auto_created=True, include_swapped=True
)
model_choices = [(
    f'{m._meta.app_label}.{m._meta.model_name}',
    f'{m._meta.app_label}.{m._meta.verbose_name}'
) for m in installed_models]


class LogConfigForm(forms.ModelForm):

    model = forms.ChoiceField(choices=model_choices)


class LogConfigFieldInLine(admin.TabularInline):

    model = models.LogConfigField


class LogConfigAdmin(admin.ModelAdmin):

    model = models.LogConfig
    list_display = ('model', 'ignore_errors', 'exclude_fields')
    search_fields = ('pk', 'model')
    list_filter = ['ignore_errors', 'exclude_fields']
    form = LogConfigForm
    inlines = [LogConfigFieldInLine]


class LogAdmin(admin.ModelAdmin):

    model = models.LogConfig
    list_display = (
        'model', 'field', 'object_id', 'log_date', 'old_value', 'new_value',
        'user', 'tenant'
    )
    search_fields = ('model', 'field', 'object_id', 'old_value')
    list_filter = ['model', 'tenant']


admin.site.register(models.LogConfig, LogConfigAdmin)
admin.site.register(models.Log, LogAdmin)



from django.db.models import Model, Field


def get_related_model(model: type[Model], rel_path: str) -> tuple[Model, list[str]]:
    names = []
    related_model = model
    for part in rel_path.split('__'):
        names.append(str(related_model._meta.verbose_name))
        try:
            next_model = related_model._meta.fields_map[part].related_model
            related_model = next_model
        except (AttributeError, KeyError):
            try:
                next_model = getattr(related_model, part).field.related_model
                if next_model is not None:
                    related_model = next_model
            except AttributeError:
                break
    names.append(str(related_model._meta.verbose_name))
    return related_model, names


def get_related_field(model: type[Model], field_path: str) -> Field:
    parts = field_path.split('__')
    rel_path = '__'.join(parts[:-1])
    rel_model, _ = get_related_model(model, rel_path)
    return rel_model._meta.get_field(parts[-1])



import logging
import re
from datetime import datetime, date, timedelta

from django.contrib.auth import get_user_model
from django.utils.translation import gettext_lazy as _
from django import template
from django.db.models import (
    Manager, DecimalField, IntegerField, FloatField, Model, Q
)
from django.apps import apps
from django.template.loader import render_to_string
from django.utils.safestring import mark_safe
from django.forms import widgets
from accrete.contrib.ui.models import Theme
from accrete.tenant import get_tenant

_logger = logging.getLogger(__name__)
register = template.Library()
User = get_user_model()


@register.simple_tag(name='combine_templates')
def combine_templates(template_name, request=None):
    html = ''
    for app in apps.app_configs.keys():
        try:
            html += render_to_string(
                f'{app}/{template_name}',
                request=request
            )
        except template.TemplateDoesNotExist:
            continue
    return mark_safe(html)


@register.filter(name='related_obj_url')
def related_obj_url(param: object, value: str):
    attr = getattr(param, value, False)
    if attr and hasattr(attr, 'get_absolute_url'):
        return attr.get_absolute_url()
    return '#'


@register.filter(name='get_attr')
def get_attr_from_string(param: object, value: str):

    def return_save(attr):
        if isinstance(attr, Manager) and hasattr(attr, 'related_model'):
            return attr.related_model
        if attr is None:
            return None
        if isinstance(attr, str):
            return attr
        return attr

    try:
        attribute = getattr(param, value, False)
    except (AttributeError,):
        _logger.warning(f'Object {param} has no attribute {value}')
        attribute = None
    return return_save(attribute)


@register.filter(name='has_attr')
def has_attr(param: object, value:str):
    return hasattr(param, value)


@register.filter(name='get_item')
def get_item(param, value):
    return param[value]


@register.filter(name='get_model')
def get_model(instance):
    return instance._meta.model


@register.filter(name='verbose_field_name')
def verbose_field_name(param: object, value: str):
    if not param:
        return ''
    if isinstance(param, Model):
        name = param._meta.get_field(value).verbose_name
    else:
        name = getattr(param.model, value).field.verbose_name
    return name


@register.filter(name='table_display')
def table_display(param: object, value: str):
    field = getattr(param, value)
    if isinstance(field, Manager):
        return mark_safe(f"""
            <span>            
                <div class="tags">
                    {''.join(['<span class="tag">{}</span>'.format(o) for o in field.all()])}
                </div>
            </span>
        """)
    return get_attr_from_string(param, value)


@register.filter(name='table_alignment')
def table_alignment(param, value):
    if isinstance(param, str):
        return 'left'
    if isinstance(param, Model):
        field = param._meta.get_field(value)
    else:
        field = getattr(param.model, value).field
    if isinstance(field, (DecimalField, IntegerField, FloatField)):
        return 'right'
    return 'left'


@register.filter(name='message_class')
def message_class(message):
    if message.level == 25:
        return 'is-success'
    if message.level == 30:
        return 'is-warning'
    if message.level == 40:
        return 'is-danger'


@register.filter(name='timedelta_cast')
def timedelta_cast(td: timedelta, code: str) -> str | None:
    if not isinstance(td, timedelta):
        return None
    codes = ['days', 'hours', 'minutes', 'seconds', 'microseconds']
    if code not in codes:
        return None
    return str(getattr(td, code))


@register.filter(name='weekday')
def datetime_to_weekday(dt: datetime|date|str, default=None) -> str:
    if dt is None:
        return default
    if isinstance(dt, str):
        dt = datetime.strptime(dt, '%Y-%m-%d')
    mapping = {
        1: _('Mon'),
        2: _('Tue'),
        3: _('Wed'),
        4: _('Thu'),
        5: _('Fri'),
        6: _('Sat'),
        7: _('Sun')
    }
    return mapping[dt.isoweekday()]


@register.filter(name='xrefsave')
def x_ref_save(param: str):
    return re.sub(r'[^A-Za-z]+','', param)


@register.filter(name='wrap_form_field')
def wrap_form_fields(field, icon=None):
    if isinstance(field.field.widget, widgets.Textarea):
        html = render_to_string(
            'ui/templatetags/field.html#textarea',
            {'field': field, 'icon': None}
        )
        return mark_safe(html)
    html = render_to_string(
        'ui/templatetags/field.html#form_field',
        {'field': field, 'icon': icon}
    )
    return mark_safe(html)


@register.filter(name='wrap_model_field')
def wrap_model_field(instance, field_name):
    field = instance._meta.get_field(field_name)
    if field.choices:
        value = getattr(instance, f'get_{field_name}_display')()
    else:
        value = getattr(instance, field_name)
    html = render_to_string(
        'ui/templatetags/field.html#model_field', {
            'label': field.help_text or field.verbose_name,
            'value': value
        })
    return mark_safe(html)


@register.filter(name='default_if_falsy')
def default_if_falsy(value, default):
    if bool(value):
        return value
    return default


@register.simple_tag(name='custom_theme')
def custom_theme(user: User) -> str:
    if user.is_anonymous:
        return ''
    tenant = get_tenant()
    tenant_theme = Theme.objects.filter(
        tenant=tenant, tenant__isnull=False
    ).first()
    if tenant_theme and tenant_theme.force_tenant_theme:
        return mark_safe(tenant_theme.theme_markup)
    if user.theme == 'custom':
        theme = Theme.objects.filter(user=user).first()
        return theme and mark_safe(theme.theme_markup) or ''
    if user.theme == 'preset' and tenant_theme:
        return mark_safe(tenant_theme.theme_markup)
    return ''



from django.db import models
from accrete.tenant import get_tenant


class TenantManager(models.Manager):

    def get_queryset(self):
        queryset = super().get_queryset()
        tenant = get_tenant()
        if tenant:
            queryset = queryset.filter(tenant=tenant)
        return queryset

    def bulk_create(
            self,
            objs,
            batch_size=None,
            ignore_conflicts=False,
            update_conflicts=False,
            update_fields=None,
            unique_fields=None,
    ):
        tenant = get_tenant()
        if tenant is None and any(obj.tenant_id in (False, None) for obj in objs):
            raise ValueError(
                'Tenant must be set for all objects when calling '
                'bulk_create without an active tenant set.'
            )
        list(map(
            lambda x: setattr(x, 'tenant_id', tenant.pk),
            filter(lambda o: o.tenant_id is None, objs)
        ))
        if tenant is not None and any(obj.tenant_id != tenant.pk for obj in objs):
            raise ValueError(
                'Objects must have set the active Tenant when calling '
                'bulk_create while a tenant is active.'
            )
        return super().bulk_create(
            objs, batch_size=batch_size, ignore_conflicts=ignore_conflicts,
            update_conflicts=update_conflicts, update_fields=update_fields,
            unique_fields=unique_fields
        )


class MemberManager(TenantManager):

    def get_queryset(self):
        queryset = super().get_queryset().select_related('tenant', 'user')
        return queryset


class AccessGroupManager(models.Manager):

    def tenant_groups(self):
        return self.get_queryset().filter(apply_on='tenant')

    def member_groups(self):
        return self.get_queryset().filter(apply_on='member')



# Generated by Django 5.2.1 on 2025-08-07 16:18

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('accrete', '0009_alter_accessgroup_name'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Theme',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('color_primary', models.CharField(max_length=7, verbose_name='Primary Color')),
                ('color_success', models.CharField(max_length=7, verbose_name='Success Color')),
                ('color_link', models.CharField(max_length=7, verbose_name='Link Color')),
                ('color_warning', models.CharField(max_length=7, verbose_name='Warning Color')),
                ('color_danger', models.CharField(max_length=7, verbose_name='Danger Color')),
                ('theme_markup', models.TextField(blank=True, null=True, verbose_name='Theme Markup')),
                ('tenant', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='tenant_themes', to='accrete.tenant', verbose_name='Tenant')),
                ('user', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='user_themes', to=settings.AUTH_USER_MODEL, verbose_name='User')),
            ],
            options={
                'verbose_name': 'Theme',
                'verbose_name_plural': 'Themes',
                'db_table': 'accrete_theme',
                'ordering': ['pk'],
                'constraints': [models.CheckConstraint(condition=models.Q(('user__isnull', True), ('tenant__isnull', True), _connector='XOR'), name='check_user_or_tenant', violation_error_message='')],
            },
        ),
    ]



import calendar
from datetime import datetime, date, timedelta
from dateutil.relativedelta import relativedelta
from django.utils import timezone


def current_year_start() -> datetime:
    now = timezone.now()
    first_day = now.replace(
        month=1, day=1, hour=0, minute=0, second=0, microsecond=0
    )
    return first_day


def current_year_end() -> datetime:
    now = timezone.now()
    last_day = now.replace(
        month=12, day=31, hour=23, minute=59, second=59, microsecond=999999
    )
    return last_day


def current_month_start() -> datetime:
    now = timezone.now()
    first_day = now.replace(
        day=1, hour=0, minute=0, second=0, microsecond=0
    )
    return first_day


def current_month_end() -> datetime:
    now = timezone.now()
    monthrange = calendar.monthrange(now.year, now.month)
    last_day = now.replace(
        day=monthrange[1], hour=23, minute=59, second=59, microsecond=999999
    )
    return last_day


def next_relative_month_start(start_date: date) -> date:
    monthrange = calendar.monthrange(start_date.year, start_date.month)
    return start_date.replace(day=monthrange[1]) + timedelta(days=1)


def next_relative_year_start(start_date: date) -> date:
    current = current_year_end().date().replace(year=start_date.year)
    return current + relativedelta(days=1)


def day_of_month_or_last(year: int, month: int, day: int) -> int:
    max_day = calendar.monthrange(year, month)[1]
    return day <= max_day and day or max_day



import json
from django import forms
from django.db import connection
from django.db.models import JSONField
from django.core import checks
from django.utils.translation import get_language, gettext_lazy as _
from django.utils.text import capfirst
from django.conf import settings


class TranslatedCharField(JSONField):

    description = _("A JSON object to store translated strings")
    default_language = settings.LANGUAGE_CODE

    def check(self, **kwargs):
        errors = super().check(**kwargs)
        errors.extend(self._check_choices())
        return errors

    def _check_choices(self):
        errors = []
        if self.choices is not None:
            errors.append(
                checks.Error(
                    "%s does not support Choices." % type(self).__name__,
                    obj=self.model,
                    id="fields.E180",
                )
            )
        return errors

    def pre_save(self, model_instance, add):
        if not model_instance.pk:
            old_val = dict()
        else:
            with connection.cursor() as cr:
                query = """SELECT %s FROM %s WHERE id = %s""" % (
                    self.attname,
                    model_instance._meta.db_table,
                    model_instance.pk
                )
                cr.execute(query)
                row = cr.fetchone()
                row = row and row[0]
                old_val = row and json.loads(row) or dict()
        language = get_language()
        new_val = getattr(model_instance, self.attname)
        if new_val in [None, False, '']:
            return {self.default_language: ''}
        if isinstance(new_val, dict):
            old_val.update(new_val)
            # new_val = new_val.get(language, '')
        else:
            old_val.update({language: new_val})
        if language != self.default_language and not old_val.get(self.default_language):
            old_val.update({self.default_language: new_val})
        return old_val

    def from_db_value(self, value, expression, connection):
        value = super().from_db_value(value, expression, connection)
        language = get_language()
        if value is None:
            return None
        if not isinstance(value, dict):
            value = dict()
        char_val = value.get(language)
        if not char_val:
            char_val = value.get(self.default_language, None)
        return char_val

    def to_python(self, value):
        if value is None:
            return None
        if isinstance(value, str):
            return value
        language = get_language()
        char_value = ''
        if isinstance(value, dict):
            char_value = value.get(language)
            if not char_value:
                char_value = value.get(self.default_language, '')
        return char_value

    def formfield(self, form_class=None, choices_form_class=None, **kwargs):
        defaults = {
            "required": not self.blank,
            "label": capfirst(self.verbose_name),
            "help_text": self.help_text,
        }
        if self.has_default():
            if callable(self.default):
                defaults["initial"] = self.default
                defaults["show_hidden_initial"] = True
            else:
                defaults["initial"] = self.get_default()
        if self.choices is not None:
            # Fields with choices get special treatment.
            include_blank = self.blank or not (
                self.has_default() or "initial" in kwargs
            )
            defaults["choices"] = self.get_choices(include_blank=include_blank)
            defaults["coerce"] = self.to_python
            if self.null:
                defaults["empty_value"] = None
            if choices_form_class is not None:
                form_class = choices_form_class
            else:
                form_class = forms.TypedChoiceField
            # Many of the subclass-specific formfield arguments (min_value,
            # max_value) don't apply for choice fields, so be sure to only pass
            # the values that TypedChoiceField will understand.
            for k in list(kwargs):
                if k not in (
                    "coerce",
                    "empty_value",
                    "choices",
                    "required",
                    "widget",
                    "label",
                    "initial",
                    "help_text",
                    "error_messages",
                    "show_hidden_initial",
                    "disabled",
                ):
                    del kwargs[k]
        defaults.update(kwargs)
        if form_class is None:
            form_class = forms.CharField
        return form_class(**defaults)



from django.contrib import admin
from .models import Country


class CountryAdmin(admin.ModelAdmin):
    model = Country
    list_display = ('name', 'iso_code_a2', 'iso_code_a3', 'vat_prefix', 'order_priority')
    search_fields = ('name', 'iso_code_a2', 'iso_code_a3', 'vat_prefix')


admin.site.register(Country, CountryAdmin)



from django.utils.deprecation import MiddlewareMixin
from django.utils import translation
from django.conf import settings


class SetLanguageMiddleware(MiddlewareMixin):
    @staticmethod
    def process_request(request):
        if hasattr(request.user, 'language_code'):
            user_language = request.user.language_code or settings.LANGUAGE_CODE
            translation.activate(user_language)



import os
from functools import wraps
from typing import Callable

from django.http import HttpResponse, HttpResponseNotFound, HttpResponseForbidden, HttpRequest
from django.contrib.auth.mixins import LoginRequiredMixin
from django.contrib.auth.views import login_required
from django.core.exceptions import ImproperlyConfigured
from django.shortcuts import redirect, get_object_or_404, resolve_url
from django.conf import settings
from accrete.models import Tenant, Member
from accrete.tenant import get_tenant, tenant_has_group, member_has_group
from accrete import config


class TenantRequiredMixin(LoginRequiredMixin):

    # Redirect to the specified url if the group check fails
    TENANT_NOT_SET_URL = None
    GROUP_NOT_SET_URL = None

    # If set, one of the supplied groups must be present on the
    # tenant or member respectively. If the list item is of type tuple,
    # all the groups in the tuple must be present.
    TENANT_GROUPS: list[str | tuple[str]] = []
    MEMBER_GROUPS: list[str | tuple[str]] = []

    def dispatch(self, request, *args, **kwargs):
        if not self.get_tenant():
            return self.handle_tenant_not_set()
        if self.request.user.is_superuser:
            return super().dispatch(request, *args, **kwargs)
        if not self.check_tenant_group():
            return self.handle_tenant_group_not_set()
        if not self.check_member_group():
            return self.handle_member_group_not_set()
        return super().dispatch(request, *args, **kwargs)

    def handle_tenant_not_set(self):
        return redirect(
            resolve_url(self.get_tenant_not_set_url())
            + f'?next={self.request.get_full_path_info()}'
        )

    def handle_tenant_group_not_set(self):
        return redirect(self.get_group_not_set_url())

    def handle_member_group_not_set(self):
        return redirect(self.get_group_not_set_url())

    def get_tenant_not_set_url(self):
        tenant_not_set_url = (
            self.TENANT_NOT_SET_URL
            or config.ACCRETE_TENANT_NOT_SET_URL
        )
        if not tenant_not_set_url:
            cls_name = self.__class__.__name__
            raise ImproperlyConfigured(
                f"{cls_name} is missing the tenant_not_set_url attribute. "
                f"Define {cls_name}.TENANT_NOT_SET_URL, "
                f"settings.ACCRETE_TENANT_NOT_SET_URL, or override "
                f"{cls_name}.get_tenant_not_set_url()."
            )
        return tenant_not_set_url

    def get_group_not_set_url(self):
        group_not_set_url = (
            self.GROUP_NOT_SET_URL
            or config.ACCRETE_GROUP_NOT_SET_URL
        )
        if not group_not_set_url:
            cls_name = self.__class__.__name__
            raise ImproperlyConfigured(
                f"{cls_name} is missing the group_not_set_url attribute. "
                f"Define {cls_name}.GROUP_NOT_SET_URL, "
                f"settings.ACCRETE_GROUP_NOT_SET_URL, or override "
                f"{cls_name}.get_group_not_set_url()."
            )
        return group_not_set_url

    def check_tenant_group(self) -> bool:
        if not self.TENANT_GROUPS:
            return True
        for group in self.TENANT_GROUPS:
            if isinstance(group, tuple) and all([tenant_has_group(g) for g in group]):
                return True
            elif tenant_has_group(group):
                return True
        return False

    def check_member_group(self) -> bool:
        if not self.MEMBER_GROUPS:
            return True
        for group in self.MEMBER_GROUPS:
            if isinstance(group, tuple) and all([member_has_group(g) for g in group]):
                return True
            elif member_has_group(group):
                return True
        return False

    @staticmethod
    def get_tenant():
        return get_tenant()


def tenant_required(
        tenant_groups: list[str | tuple[str]] = None,
        member_groups: list[str | tuple[str]] = None,
        group_missing_action: str | Callable[[
            HttpRequest, list[str | tuple[str]], list[str | tuple[str]]
        ], HttpResponse] = None,
        redirect_field_name: str = None,
        login_url: str = None
):
    def decorator(f):
        @wraps(f)
        @login_required(
            redirect_field_name=redirect_field_name,
            login_url=login_url
        )
        def _wrapped_view(request, *args, **kwargs):

            def handle_group_missing():
                if callable(group_missing_action):
                    return group_missing_action(
                        request, tenant_groups, member_groups
                    )
                return (
                    redirect(config.ACCRETE_GROUP_NOT_SET_URL)
                    if config.ACCRETE_GROUP_NOT_SET_URL
                    else HttpResponseForbidden()
                )

            tenant = request.tenant
            if not tenant:
                return redirect(config.ACCRETE_TENANT_NOT_SET_URL)
            for tenant_group in (tenant_groups or []):
                if isinstance(tenant_group, tuple) and all([
                    tenant_has_group(g) for g in tenant_group
                ]):
                    break
                elif isinstance(tenant_group, str) and tenant_has_group(tenant_group):
                    break
                return handle_group_missing()
            for member_group in (member_groups or []):
                if isinstance(member_group, tuple) and all([
                    member_has_group(g) for g in member_group
                ]):
                    break
                elif isinstance(member_group, str) and member_has_group(member_group):
                    break
                return handle_group_missing()
            return f(request, *args, **kwargs)
        return _wrapped_view
    return decorator


@tenant_required()
def get_tenant_file(request, tenant_id, filepath):
    tenant = get_object_or_404(Tenant, pk=tenant_id)
    if not request.user.is_staff:
        member = Member.objects.filter(user=request.user, tenant=tenant)
        if not member.exists():
            return HttpResponseNotFound()
    filepath = f'{settings.MEDIA_ROOT}/{tenant_id}/{filepath}'
    if not os.path.exists(filepath):
        return HttpResponseNotFound()
    with open(filepath, 'rb') as f:
        return HttpResponse(f)



# Generated by Django 5.1.7 on 2025-03-17 18:09

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('auth', '0012_alter_user_first_name_max_length'),
        ('user', '0005_remove_user_email_or_login_set_and_more'),
    ]

    operations = [
        migrations.RemoveConstraint(
            model_name='user',
            name='email_or_login_set',
        ),
        migrations.AddField(
            model_name='user',
            name='is_managed',
            field=models.BooleanField(default=False, help_text='User with restricted functionality.', verbose_name='Is Managed'),
        ),
        migrations.AddConstraint(
            model_name='user',
            constraint=models.CheckConstraint(condition=models.Q(('email__isnull', False), ('login__isnull', False), _connector='OR'), name='email_or_login_set', violation_error_message='E-Mail or Login must be set'),
        ),
        migrations.AddConstraint(
            model_name='user',
            constraint=models.CheckConstraint(condition=models.Q(models.Q(('email__isnull', False), ('is_managed', False)), models.Q(('email__isnull', True), ('is_managed', True)), _connector='OR'), name='no_email_for_managed_user', violation_error_message='Managed users must not have an E-Mail address'),
        ),
    ]



from . import widgets
from .filter import Filter
from .response import (
    Response,
    WindowResponse,
    ListResponse,
    ListEntryResponse,
    TableResponse,
    TableRowResponse,
    DetailResponse,
    ModalResponse,
    OobResponse,
    TriggerResponse,
    ClientTrigger,
    search_select_response,
    message_response,
    add_trigger,
    update
)



# Generated by Django 5.1.4 on 2025-02-24 17:54

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('accrete', '0003_remove_member_name'),
    ]

    operations = [
        migrations.RenameModel(
            old_name='AccessGroupMember',
            new_name='MemberAccessGroupRel',
        ),
        migrations.RenameModel(
            old_name='TenantAccessGroup',
            new_name='TenantAccessGroupRel',
        ),
        migrations.AlterModelOptions(
            name='memberaccessgrouprel',
            options={'ordering': ['member'], 'verbose_name': 'Member Access Group Relation', 'verbose_name_plural': 'Member Access Group Relations'},
        ),
        migrations.AlterModelOptions(
            name='tenantaccessgrouprel',
            options={'ordering': ['tenant'], 'verbose_name': 'Tenant Access Group Relation', 'verbose_name_plural': 'Tenant Access Groups Relations'},
        ),
        migrations.RemoveField(
            model_name='accessgroup',
            name='is_public',
        ),
        migrations.AlterField(
            model_name='accessgroup',
            name='code',
            field=models.CharField(max_length=100, verbose_name='Code'),
        ),
        migrations.AlterModelTable(
            name='memberaccessgrouprel',
            table='accrete_member_access_group_rel',
        ),
    ]



# Generated by Django 5.2.1 on 2025-08-08 06:13

from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('accrete', '0009_alter_accessgroup_name'),
        ('ui', '0002_alter_theme_color_danger_alter_theme_color_link_and_more'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.AlterConstraint(
            model_name='theme',
            name='check_user_or_tenant',
            constraint=models.CheckConstraint(condition=models.Q(('user__isnull', True), ('tenant__isnull', True), _connector='XOR'), name='check_user_or_tenant', violation_error_message='A theme must be assigned either to a user or a tenant.'),
        ),
    ]



# Generated by Django 5.2.1 on 2025-07-31 05:12

import json
import accrete.fields
from django.db import migrations
from django.conf import settings


def char_to_translated_char(apps, schema):
    AccessGroup = apps.get_model('accrete', 'AccessGroup')
    default_lang = getattr(settings, 'LANGUAGE_CODE', 'en-us')
    for group in AccessGroup.objects.all():
        group.name = json.dumps({default_lang: group.name})
        group.save()


class Migration(migrations.Migration):

    dependencies = [
        ('accrete', '0008_alter_member_access_groups_and_more'),
    ]

    operations = [
        migrations.RunPython(char_to_translated_char),
        migrations.AlterField(
            model_name='accessgroup',
            name='name',
            field=accrete.fields.TranslatedCharField(verbose_name='Name'),
        ),
    ]



from django.contrib.auth.forms import AuthenticationForm
from django.contrib.auth import views, update_session_auth_hash
from django.contrib.auth.decorators import login_required
from django.http import HttpResponseForbidden, HttpResponseRedirect
from django.shortcuts import redirect, render, resolve_url
from django.utils.translation import gettext_lazy as _
from django.conf import settings

from accrete.utils import save_form
from accrete.contrib import ui
from accrete.contrib.ui.models import Theme
from accrete.contrib.ui.forms import ThemeForm
from .forms import UserForm, ChangePasswordForm, ChangeEmailForm


class LoginView(views.LoginView):

    form_class = AuthenticationForm
    template_name = 'user/login.html'
    redirect_authenticated_user = True

    def form_invalid(self, form):
        user = form.get_user()
        if user is not None and not user.is_active:
            ctx = {'to_confirm': True}
            if self.extra_context:
                self.extra_context.update(ctx)
            else:
                self.extra_context = ctx
        return super().form_invalid(form)


class LogoutView(views.LogoutView):

    def post(self, request, *args, **kwargs):
        request.skip_htmx_redirect_middleware = True
        res = super().post(request, *args, **kwargs)
        if isinstance(res, HttpResponseRedirect):
            res.headers['HX-Redirect'] = res.url
            res.status_code = 200
        return res

    def get_success_url(self):
        return resolve_url(settings.LOGIN_URL)


@login_required()
def user_detail(request):
    form = UserForm(
        initial={'language_code': request.user.language_code},
        instance=request.user
    )
    theme = Theme.objects.filter(user=request.user).first()
    theme_form = ThemeForm(instance=theme, prefix='theme', initial={'user': request.user})
    refresh = False
    if request.method == 'POST':
        form = save_form(UserForm(request.POST, instance=request.user))
        theme_form = save_form(ThemeForm(request.POST, instance=theme, prefix='theme', initial={'user': request.user}))
        if (form.is_saved or not form.has_changed()) or (theme_form.is_saved or not theme_form.has_changed()):
            refresh = True
    res = ui.WindowResponse(
        title=str(_('User Preferences')),
        overview_template='user/user_preferences.html#data',
        header_template='user/user_preferences.html#header',
        context=dict(user=request.user, form=form, theme_form=theme_form),
        is_centered=True
    ).response(request, replace_body=False)
    if refresh:
        res.headers['HX-Refresh'] = 'true'
    return res


@login_required()
def user_change_password(request):
    if request.user.is_managed:
        return HttpResponseForbidden()
    form = ChangePasswordForm(instance=request.user)
    update = bool(request.method == 'POST')
    if request.method == 'POST':
        form = save_form(ChangePasswordForm(request.POST, instance=request.user))
        if form.is_saved:
            update_session_auth_hash(request, form.instance)
            return redirect(
                resolve_url('user:detail')
                + f'?{request.GET.urlencode()}'
            )
    return ui.ModalResponse(
        title=str(_('Change Password')),
        modal_id='change-password',
        template='user/change_password.html',
        is_update=update,
        context=dict(form=form, user=request.user)
    ).response(request)


@login_required()
def user_change_email(request):
    if request.user.is_managed:
        return HttpResponseForbidden()
    form = ChangeEmailForm(instance=request.user)
    update = bool(request.method == 'POST')
    if request.method == 'POST':
        form = save_form(ChangeEmailForm(request.POST, instance=request.user))
        if form.is_saved:
            return redirect(
                resolve_url('user:detail')
                + f'?{request.GET.urlencode()}'
            )
    return ui.ModalResponse(
        title=str(_('Change E-Mail')),
        modal_id='change-email',
        template='user/change_email.html',
        is_update=update,
        context=dict(form=form, user=request.user)
    ).response(request)


def password_forgotten(request):
    return render(request, 'user/')


from django.contrib import admin
from django import forms
from .models import User


class UserForm(forms.ModelForm):

    login = forms.CharField(
        required=False
    )

    email = forms.EmailField(
        required=False
    )


class UserAdmin(admin.ModelAdmin):

    model = User
    form = UserForm
    list_display = (
        'username',
        'email',
        'login',
        'first_name',
        'last_name'
    )
    search_fields = [
        'email',
        'login',
        'username',
        'first_name',
        'last_name'
    ]
    list_filter = [
        'is_superuser',
        'is_staff',
        'is_active',
        'is_managed'
    ]


admin.site.register(User, UserAdmin)



# Generated by Django 5.1.4 on 2025-02-19 18:25

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('country', '0002_auto_20250219_1759'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='country',
            options={'ordering': ['-order_priority', 'name'], 'verbose_name': 'Country', 'verbose_name_plural': 'Countries'},
        ),
    ]
